CONFIG = dict((('extra_compile_args', ['-UNDEBUG', '-w']), ('extra_sources', []), ('extra_library_dirs', ['/usr/local/lib']), ('extra_include_dirs', ['/usr/local/include', 'lib/py-lmdb']), ('libraries', ['lmdb'])))

